from embedbase.database import VectorDatabase


class Weaviate(VectorDatabase):
    def __init__(
        self,
    ):
        """
        weaviate.io
        """
        raise NotImplementedError(
            "Weaviate is not supported."
            + "If you want to use Weaviate, please"
            + " create an issue "
            + "https://github.com/different-ai/embedbase/issues/new/choose"
        )
